/***************************************************************************
                          deviceparser.cpp  -  description
                             -------------------
    begin                : mar dc 20 2005
    copyright            : (C) 2005 by Gibaud Alain
    email                : alain.gibaud@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "deviceparser.h"
#include "device.h"
#include "devicedata.h"

#include <ctype.h>
#include <sstream>

using namespace std ;

// ---------------------------------------------------
deviceParser::deviceParser(device *d, bool v)
{
    kbuffer = 0 ;
    line_number = 0 ;
    dev = d ;
    verbose = v ;
}

// ---------------------------------------------------
deviceParser::~deviceParser()
{
}

// ---------------------------------------------------
bool deviceParser::getLine()
{
    kbuffer = 0 ;
    buffer.clear() ;
    bool stat ;

    do
    {
        if( (stat = getline(stream, buffer)) )
        {
            // cope with DOS CR-LF
            if( buffer[buffer.size()-1] == '\r' )
                buffer[buffer.size()-1] = '\n' ;
            else
                buffer += '\n' ;

            ++line_number ;

            // replace tabs with 8 blanks
            string::size_type p ;
            while( (p =  buffer.find('\t')) != string::npos)
            {
                buffer.replace(p,1,"        ") ;
            }
        }
    }
    while( stat && isBlank() ) ;

    return stat ;
}

// ---------------------------------------------------
/** return next char, but keep it into buffer
buffer is filled if needed and not EOF
*/
char deviceParser::ch()
{
    if( kbuffer < buffer.size() )
        return  buffer[kbuffer] ;
    else   if( getLine())
        return ch() ;
    else
        return EOF ;
}
// ---------------------------------------------------
/** get next char.
 */
char deviceParser::gch()
{
    char c = ch() ;
    if( c != 0 ) ++kbuffer ;
    return c ;
}
// ---------------------------------------------------
bool deviceParser::getDeviceName(string& name)
{
    while( getLine() )
    {
        skipWS() ;
        if(match("#") && skipWS() && match("device=") && ident(name)) return true ;
    }
    return false ;
}


// ---------------------------------------------------

/** No descriptions */
string deviceParser::parseDevice(const string&  fname)
{
    string status ;

    stream.open( fname.c_str() ) ;
    if( !stream.is_open()) return "File not found : "+ fname ;

    filename = fname ;
    string devicename ;

    // compute device name from file name
    devicename = fname ;
    string::size_type i = devicename.rfind(".") ;
    if( i != string::npos) devicename.erase(i) ;
    i = devicename.rfind("/") ;
    if( i != string::npos) devicename.erase(0,i+1) ;
    for( i = 0 ; i < devicename.size() ; ++i)
       devicename[i] = toupper(devicename[i]) ;
    

    // dev = new device ;
    dev->name = devicename ;


    if( (status =  parseAllEntries(dev->data)) != "")
    {
        stream.close() ; return status ;
    }



    stream.close() ;
    return "" ;
}



// ---------------------------------------------------
bool deviceParser::skipWS()
{
    while( ch() == ' ' ) gch()  ;
    return true ;
}


// ---------------------------------------------------
bool deviceParser::match(const string& id)
{
    unsigned int ksav = kbuffer ;
    for(unsigned int i = 0 ; i < id.size() ; ++i, gch() )
    {
        if( ch() == '\n' || ch() != id[i] )
        {
            kbuffer = ksav ;
            return false ;
        }
    }
    return true ;
}
// ---------------------------------------------------
/** No descriptions */
bool deviceParser::ident(string& id)
{
    char c ;
    id.clear() ;

    if( !isalpha(ch())) return false ;
    id += gch() ;

    for( ; (c=ch()) && isalnum(c) ; gch() )
    {
        id += c ;
    }
    return id.size() != 0 ;
}

// ---------------------------------------------------
/** parse level lev, and add them to data vector*/
string deviceParser::parseAllEntries(vector<deviceData*>& data)
{
    string status ;
    int curlevel = 0, newlevel ;

    while( !dds.empty() ) dds.pop() ;
    
    dds.push(&data) ;
    
    while(status.empty() && getLine())
    {
        if( isComment() || isBlank() )  continue ;

        if( levelPrefix(newlevel) )
        {
          if( newlevel == curlevel)
            status = parseNewData( *dds.top() ) ;
          else if( newlevel == curlevel+1)
          {
            dds.push( &(dds.top()->back()->subdata) ) ;
            status = parseNewData( *dds.top() ) ;
            curlevel = newlevel ;
          }
          else if( newlevel < curlevel)
          {
            // return to proper level
            while( curlevel > newlevel && !dds.empty() )
            {
              dds.pop() ;
              --curlevel ;
            }
            if( curlevel != newlevel )
              status = ln() + "parseEntry: internal error - stack empty" ;
            else
              status = parseNewData( *dds.top() ) ;
          }
          else
            status =  ln() + "parseEntry: too deeply indented (please fix)" ;
        }
        else
          status = ln() + "parseEntry: malformed indentation (please fix)" ;        
    }
    return status ;
}
// ---------------------------------------------------
// ---------------------------------------------------
/** True if we are at Beginning Of Line
and if there is something to read
*/
bool deviceParser::BOL()
{
    return (kbuffer == 0) && stream ;
}
/** search for series of 4 white-spaces.
ie: 0, 4, 8, ..., 4*k white-spaces

return true if such a serie is found, and update k accordingly.
if 4*k serie not found, k is set to the number of blanks found
*/
bool deviceParser::levelPrefix(int& k)
{
    for( k = 0 ; ch() == ' '  ; gch()) ++k ;

    if( k % 4 == 0)
    {
        k /= 4 ;
        return true ;
    }
    else if( k % 4 == 1)
    {
      // assume one char indent error - autofix
      --k ;
        if( verbose) cout << ln() << "indent autofix: " << (k+1) << " to " << k << endl ;
        k /= 4 ;
        return true ;
    }
    else if( k % 4 == 3)
    {
      // assume one char indent error - autofix
        ++k ;
        if(verbose) cout << ln() << "indent autofix: " << (k-1) << " to " << k << endl ;
        k /= 4 ;
        return true ;
    }
    return false ;
}
/** try to move cursor backward to n chars
 */
bool deviceParser::pushBack(unsigned int n)
{
    bool r = true ;
    if( kbuffer >= n ) kbuffer -= n ;
    else r = false ;
    return r ;
}
/** No descriptions */
string deviceParser::parseNewData(vector<deviceData*>&  data)
{
    string id,r ;

    if( ident(id)  )
    {
        deviceData *d = new deviceData(id) ;
        if( (r=parseParameters(d->params)) == ""  )
        {
            data.push_back(d) ;
        }
        else
        {
            delete d ;
        }

    }
    else
        r = ln() + "Can't decode data" ;

    // skip remaining of line
    //if( !BOL()) getLine() ;
    return r ;
}


/** Parse a serie of comma separated  'parameter=value' pairs. */
string deviceParser::parseParameters(map<string, string>& table)
{
    string r, key, value  ;
    char delim ;

    skipWS() ;

    if( ch() == '=')
    {
        gch() ;
        return parseUniqueParameter(table) ;
    }


    if( ch() != '(' )
    {
        r = ln() + "Missing '(' at beginning of parameter list" ;
        return r ;
    }
    gch() ; skipWS() ;

    if( !isalpha(ch()) )
    {
        // unique parameter : no key
        r = parseUniqueRange(table) ;
    }
    else
    {
        while ( ch() != '\n' && ch() != ')')
        {

            if( ident(key) && match("=") )
            {
                value.clear() ;
                if( ch() == '"' || ch() == '\'')
                {
                    delim = gch() ;
                    while( ch() != delim && ch() != '\n' )
                        value += gch() ;
                    gch() ;
                }
                else
                {
                    while( ch() != ' ' && ch() != '\n' && ch() != ')' )
                        value += gch() ;
                }
            }
            else
            {
                r = ln() + "Incorrect key=value syntax in parameter list" ;
                return r ;
            }
            skipWS() ;
            table[key] = value ;
        }
    }



    if( ch() != ')' )
    {
        r = ln() + "Missing ')' at end of parameter list" ;
        return r ;
    }
    else gch() ;
    return r ;
}
/** return current line number as a string */
string deviceParser::ln()
{
    ostringstream s ;
    s << '(' << line_number  << "-" << buffer[kbuffer] << ") " << buffer   ;

    return s.str() ;
}
/** No descriptions */
string deviceParser::parseUniqueParameter(map<string, string>& table)
{
    string value ,r ;
    for( ; ch() != '\n' && ch() != ' '  ; )
        value += gch() ;
    table["1"] = value ;

    if( value == "" ) ln() + "Empty single parameter" ;
    return r ;
}
/** No descriptions */
string deviceParser::parseUniqueRange(map<string, string>& table)
{
    string value ,r ;

    for( ; ch() != '\n' && ch() != ')'  ; )
        value += gch() ;
    if( value == "")
        r = ln() + "Empty range parameter" ;
    else
        table["1"] = value ;
    return r ;
}
/** No descriptions */
bool deviceParser::isComment()
{
    char c ;
    while((c=ch()) == ' ' && c != '\n' && c != EOF ) gch() ;
    kbuffer = 0 ;
    return c == '#' ;
}
/** true if line contains only blanks or tabs before newline */
bool deviceParser::isBlank()
{
    char c ;
    while((ch() != '\n' && ch() != EOF))
    {
        c = gch() ;
        if( c != ' ' && c != '\t' )
        {
            kbuffer = 0 ;
            return 0 ;
        }
    }
    kbuffer = 0 ;
    return 1 ;
}
